import os
import json

from taskflows import OpenChemistryTaskFlow

class ChemmlTaskFlow(OpenChemistryTaskFlow):

    @property
    def code_label(self):
        return 'chemml'

    @property
    def docker_image(self):
        return 'openchemistry/chemml:latest'

    def input_generator(self, params, cjson, tmp_file):
        tmp_file.write(json.dumps(cjson).encode())

    def select_output_files(self, filenames):
        do_copy = [False] * len(filenames)
        for i, file in enumerate(filenames):
            if file.endswith('.out'):
                do_copy[i] = True
        return do_copy



import json

from .utils import _cclib_to_cjson_basis, _cclib_to_cjson_mocoeffs, _cclib_to_cjson_vibdisps

from .base import BaseReader

class Psi4Reader(BaseReader):

    def read(self):
        import cclib

        data = cclib.io.ccread(self._file)
        cjson = json.loads(cclib.ccwrite(data, outputtype='cjson',))

        # The cjson produced by cclib is not directly usable in our platform
        # Basis, moCoefficients, and normal modes need to be further converted

        # Cleanup original cjson
        if 'orbitals' in cjson['atoms']:
            del cjson['atoms']['orbitals']
        if 'properties' in cjson:
            del cjson['properties']
        if 'vibrations' in cjson:
            del cjson['vibrations']
        if 'optimization' in cjson:
            del cjson['optimization']
        if 'diagram' in cjson:
            del cjson['diagram']
        if 'inchi' in cjson:
            del cjson['inchi']
        if 'inchikey' in cjson:
            del cjson['inchikey']
        if 'smiles' in cjson:
            del cjson['smiles']

        # Convert basis set info
        if hasattr(data, 'gbasis'):
            basis = _cclib_to_cjson_basis(data.gbasis)
            cjson['basisSet'] = basis

        # Convert mo coefficients
        if hasattr(data, 'mocoeffs'):
            mocoeffs = _cclib_to_cjson_mocoeffs(data.mocoeffs)
            cjson.setdefault('orbitals', {})['moCoefficients'] = mocoeffs
        
        # Convert mo energies
        if hasattr(data, 'moenergies'):
            moenergies = list(data.moenergies[-1])
            cjson.setdefault('orbitals', {})['energies'] = moenergies

        if hasattr(data, 'nelectrons'):
            cjson.setdefault('orbitals', {})['electronCount'] = int(data.nelectrons)
        
        if hasattr(data, 'homos') and hasattr(data, 'nmo'):
            homos = data.homos
            nmo = data.nmo
            if len(homos) == 1:
                occupations = [2 if i <= homos[0] else 0 for i in range(nmo)]
                cjson.setdefault('orbitals', {})['occupations'] = occupations

        # Convert normal modes
        if hasattr(data, 'vibfreqs'):
            vibfreqs = list(data.vibfreqs)
            cjson.setdefault('vibrations', {})['frequencies'] = vibfreqs

        if hasattr(data, 'vibdisps'):
            vibdisps = _cclib_to_cjson_vibdisps(data.vibdisps)
            cjson.setdefault('vibrations', {})['eigenVectors'] = vibdisps

        # Add a placeholder intensities array
        if 'vibrations' in cjson and 'frequencies' in cjson['vibrations']:
            if 'intensities'  not in cjson['vibrations']:
                cjson['vibrations']['intensities'] = [1 for i in range(len(cjson['vibrations']['frequencies']))]
            if 'modes'  not in cjson['vibrations']:
                cjson['vibrations']['modes'] = [i + 1 for i in range(len(cjson['vibrations']['frequencies']))]

        
        # Convert calculation metadata
        if hasattr(data, 'metadata'):
            metadata = data.metadata
            if 'basis_set' in metadata:
                cjson.setdefault('metadata', {})['basisSet'] = metadata['basis_set'].lower()
            if 'functional' in metadata:
                cjson.setdefault('metadata', {})['functional'] = metadata['functional'].lower()
            if 'methods' in metadata and len(metadata['methods']) > 0:
                cjson.setdefault('metadata', {})['theory'] = metadata['methods'][0].lower()

        return cjson



from IPython.display import display, JSON, DisplayObject

# A display class that can be used within a notebook.
#   from openchemistry.notebook import CJSON
#   CJSON(data)

DEFAULT_ISO = 0.05;
DEFAULT_ISO_SURFACES = [{
    'value': DEFAULT_ISO,
    'color': 'blue',
    'opacity': 0.9,
  }, {
    'value': -DEFAULT_ISO,
    'color': 'red',
    'opacity': 0.9
  }]

class CJSON(JSON):
    """A display class for displaying CJSON visualizations in the Jupyter Notebook and IPython kernel.
    CJSON expects a JSON-able dict, not serialized JSON strings.
    Scalar types (None, number, string) are not allowed, only dict containers.
    """



    def __init__(self, data=None, url=None, filename=None, vibrational=True, structure=True,
                 iso_surfaces=DEFAULT_ISO_SURFACES, animate_mode=None, calculation_id=None,
                 iso_value=DEFAULT_ISO, mo=None):
        super(CJSON, self).__init__(data, url, filename)
        self.metadata['vibrational'] = vibrational
        self.metadata['structure'] = structure
        self.metadata['isoValue'] = iso_value
        self.metadata['isoSurfaces'] = iso_surfaces
        self.metadata['animateMode'] = animate_mode
        self.metadata['mo'] = mo
        self.metadata['calculationId'] = calculation_id

    def _ipython_display_(self):
        bundle = {
            'application/vnd.oc.cjson+json': self.data,
            'text/plain': '<jupyterlab_cjson.CJSON object>'
        }
        metadata = {
            'application/vnd.oc.cjson+json': self.metadata
        }
        display(bundle, metadata=metadata, raw=True)

class FreeEnergy(JSON):
    """A display class for displaying free energy visualizations in the Jupyter Notebook and IPython kernel.
    FreeEnergy expects a JSON-able dict.
    """

    def __init__(self, data=None, url=None, filename=None):
        super(FreeEnergy, self).__init__(data, url, filename)

    def _ipython_display_(self):
        bundle = {
            'application/vnd.oc.free_energy+json': self.data,
            'text/plain': '<jupyterlab_cjson.FreeEnergy object>'
        }
        metadata = {
            'application/vnd.oc.free_energy+json': self.metadata
        }
        display(bundle, metadata=metadata, raw=True)

class CalculationMonitor(DisplayObject):
    """
    A display class for monitoring calculations Jupyter Notebook and IPython kernel.
    """

    def __init__(self, data=None, url=None, filename=None):
        super(CalculationMonitor, self).__init__(data, url, filename)

    def _ipython_display_(self):
        bundle = {
            'application/vnd.oc.calculation+json': self.data,
            'text/plain': '<jupyterlab_cjson.CalculationMonitor object>'
        }
        metadata = {
            'application/vnd.oc.calculation+json': {}
        }
        display(bundle, metadata=metadata, raw=True)

    def __getattr__(self, name):
        # This is a little fragile, it seem that ipython is looking for the
        # absence of _ipython_canary_method_should_not_exist_, so only return
        # self for 'public' methods.
        if name[0] != '_':
            return self
        else:
            return DisplayObject.__getattr__(self, name)

    def __call__(self):
        return self
